"""Download and process podcast feeds and media files."""
# and return xml that can be served to download them

import contextlib
import datetime
import re
import time
from http import HTTPStatus
from typing import TYPE_CHECKING

import aiohttp
from lxml import etree

from archivepodcast.config import AppConfig, PodcastConfig
from archivepodcast.instances.health import health
from archivepodcast.utils.log_messages import log_aiohttp_exception
from archivepodcast.utils.logger import get_logger
from archivepodcast.utils.rss import tree_no_episodes
from archivepodcast.utils.time import warn_if_too_long

from .asset_downloader import AssetDownloader
from .constants import AUDIO_FORMATS, DOWNLOAD_RETRY_COUNT, IMAGE_FORMATS
from .helpers import delay_download

if TYPE_CHECKING:
    from archivepodcast.config import AppConfig, PodcastConfig  # pragma: no cover
else:
    AppConfig = object
    PodcastConfig = object

logger = get_logger(__name__)


# These make the name spaces appear nicer in the generated XML
etree.register_namespace("googleplay", "http://www.google.com/schemas/play-podcasts/1.0")
etree.register_namespace("atom", "http://www.w3.org/2005/Atom")
etree.register_namespace("podcast", "https://podcastindex.org/namespace/1.0")
etree.register_namespace("itunes", "http://www.itunes.com/dtds/podcast-1.0.dtd")
etree.register_namespace("media", "http://search.yahoo.com/mrss/")
etree.register_namespace("sy", "http://purl.org/rss/1.0/modules/syndication/")
etree.register_namespace("content", "http://purl.org/rss/1.0/modules/content/")
etree.register_namespace("wfw", "http://wellformedweb.org/CommentAPI/")
etree.register_namespace("dc", "http://purl.org/dc/elements/1.1/")
etree.register_namespace("slash", "http://purl.org/rss/1.0/modules/slash/")
etree.register_namespace("rawvoice", "http://www.rawvoice.com/rawvoiceRssModule/")
etree.register_namespace("spotify", "http://www.spotify.com/ns/rss/")
etree.register_namespace("feedburner", "http://rssnamespace.org/feedburner/ext/1.0")


class PodcastsDownloader(AssetDownloader):
    """PodcastDownloader object."""

    async def download_podcast(
        self,
    ) -> etree._ElementTree | None:
        """Parse the rss, Download all the assets, this is main."""
        self._feed_download_healthy = True
        feed_rss_healthy = True
        tree = await self._download_and_parse_rss()

        if tree:
            if tree_no_episodes(tree):
                logger.error(
                    "Downloaded podcast rss %s has no episodes, not writing to disk", self._podcast.name_one_word
                )
                feed_rss_healthy = False
            else:
                # Write rss to disk
                tree.write(
                    str(self._rss_file_path),
                    encoding="utf-8",
                    xml_declaration=True,
                )
                logger.debug("[%s] Wrote rss to disk: %s", self._podcast.name_one_word, self._rss_file_path)
        else:
            feed_rss_healthy = False
            logger.error("Unable to download podcast, something is wrong, will try to load from file")

        if not feed_rss_healthy:
            self._feed_download_healthy = False

        health.update_podcast_status(
            self._podcast.name_one_word,
            healthy_feed=feed_rss_healthy,
            healthy_download=self._feed_download_healthy,
        )

        return tree

    async def _download_and_parse_rss(self) -> etree._ElementTree | None:
        """Download and parse the podcast RSS feed."""
        content = None
        for n in range(DOWNLOAD_RETRY_COUNT):
            start_time = time.time()
            content, status = await self._fetch_podcast_rss()
            warn_if_too_long(f"[{self._podcast.name_one_word}] download podcast rss", time.time() - start_time)

            if status in {HTTPStatus.NOT_FOUND, HTTPStatus.FORBIDDEN}:
                logger.error(
                    "[%s] RSS download attempt failed with HTTP status %s, not retrying",
                    self._podcast.name_one_word,
                    status,
                )
                return None
            if status not in {HTTPStatus.OK, HTTPStatus.MOVED_PERMANENTLY, HTTPStatus.FOUND}:
                logger.warning(
                    "[%s] RSS download attempt %d/%d failed with HTTP status %s",
                    self._podcast.name_one_word,
                    n + 1,
                    DOWNLOAD_RETRY_COUNT,
                    status,
                )
            if content is not None:
                break
            await delay_download(n)

        if content is None:
            return None

        logger.debug("[%s] Success fetching podcast RSS", self._podcast.name_one_word)

        try:
            podcast_rss = etree.fromstring(content)
        except etree.XMLSyntaxError:
            logger.error(  # noqa: TRY400
                "[%s] Downloaded podcast rss (length %d) is not valid XML, cannot process podcast feed",
                self._podcast.name_one_word,
                len(content),
            )
            self._feed_download_healthy = False
            return None
        podcast_rss = etree.fromstring(content)
        logger.debug("[%s] Downloaded rss feed, processing", self._podcast.name_one_word)
        logger.trace(str(podcast_rss))

        xml_first_child = podcast_rss[0]
        await self._process_podcast_rss(xml_first_child)
        podcast_rss[0] = xml_first_child

        return etree.ElementTree(podcast_rss)

    async def _fetch_podcast_rss(self) -> tuple[bytes | None, HTTPStatus | None]:
        """Fetch the podcast RSS feed."""
        logger.debug(
            "[%s] Starting fetch for podcast RSS: %s", self._podcast.name_one_word, self._podcast.url.encoded_string()
        )
        try:
            async with self._aiohttp_session.get(self._podcast.url.encoded_string()) as response:
                return await response.read(), HTTPStatus(response.status)

        except aiohttp.ClientError as e:
            log_aiohttp_exception(self._podcast.name_one_word, self._podcast.url.encoded_string(), e, logger)
        return None, None

    # region RSS Hell

    async def _process_podcast_rss(self, xml_first_child: etree._Element) -> None:
        """Process the podcast rss and update it with new values."""
        for channel in xml_first_child:
            await self._process_channel_tag(channel)

    async def _process_channel_tag(self, channel: etree._Element) -> None:  # noqa: C901 # There is no way to avoid this really, there are many tag types
        """Process individual channel tags in the podcast rss."""
        match channel.tag:
            case "link":
                self._handle_link_tag(channel)
            case "title":
                self._handle_title_tag(channel)
            case "description":
                self._handle_description_tag(channel)
            case "{http://www.w3.org/2005/Atom}link":
                self._handle_atom_link_tag(channel)
            case "{http://www.itunes.com/dtds/podcast-1.0.dtd}owner":
                self._handle_itunes_owner_tag(channel)
            case "{http://www.itunes.com/dtds/podcast-1.0.dtd}author":
                self._handle_itunes_author_tag(channel)
            case "{http://www.itunes.com/dtds/podcast-1.0.dtd}new-feed-url":
                self._handle_itunes_new_feed_url_tag(channel)
            case "{http://www.itunes.com/dtds/podcast-1.0.dtd}image":
                await self._handle_itunes_image_tag(channel)
            case "image":
                await self._handle_image_tag(channel)
            case "item":
                await self._handle_item_tag(channel)
            case _:
                logger.trace(
                    "[%s] Unhandled root-level XML tag %s, (under channel.tag) leaving as-is",
                    self._podcast.name_one_word,
                    channel.tag,
                )

    def _handle_link_tag(self, channel: etree._Element) -> None:
        """Handle the link tag in the podcast rss."""
        logger.trace("[%s] Podcast link: %s", self._podcast.name_one_word, str(channel.text))
        channel.text = self._app_config.inet_path.encoded_string()

    def _handle_title_tag(self, channel: etree._Element) -> None:
        """Handle the title tag in the podcast rss."""
        logger.debug("[%s] Source Podcast title: %s", self._podcast.name_one_word, str(channel.text))
        if self._podcast.new_name != "":
            channel.text = self._podcast.new_name

    def _handle_description_tag(self, channel: etree._Element) -> None:
        """Handle the description tag in the podcast rss."""
        logger.trace("[%s] Podcast description: %s", self._podcast.name_one_word, str(channel.text))
        channel.text = self._podcast.description

    def _handle_atom_link_tag(self, channel: etree._Element) -> None:
        """Handle the Atom link tag in the podcast rss."""
        logger.trace("[%s] Atom link: %s", self._podcast.name_one_word, str(channel.attrib["href"]))
        channel.attrib["href"] = self._app_config.inet_path.encoded_string() + "rss/" + self._podcast.name_one_word
        channel.text = " "

    def _handle_itunes_owner_tag(self, channel: etree._Element) -> None:
        """Handle the iTunes owner tag in the podcast rss."""
        logger.trace("[%s] iTunes owner: %s", self._podcast.name_one_word, str(channel.text))
        for child in channel:
            if child.tag == "{http://www.itunes.com/dtds/podcast-1.0.dtd}name":
                if self._podcast.new_name == "":
                    self._podcast.new_name = child.text or ""
                child.text = self._podcast.new_name
            if child.tag == "{http://www.itunes.com/dtds/podcast-1.0.dtd}email":
                if self._podcast.contact_email == "":
                    self._podcast.contact_email = child.text or ""
                child.text = self._podcast.contact_email

    def _handle_itunes_author_tag(self, channel: etree._Element) -> None:
        """Handle the iTunes author tag in the podcast rss."""
        logger.trace("[%s] iTunes author: %s", self._podcast.name_one_word, str(channel.text))
        if self._podcast.new_name != "":
            channel.text = self._podcast.new_name

    def _handle_itunes_new_feed_url_tag(self, channel: etree._Element) -> None:
        """Handle the iTunes new-feed-url tag in the podcast rss."""
        logger.trace("[%s] iTunes new-feed-url: %s", self._podcast.name_one_word, str(channel.text))
        channel.text = self._app_config.inet_path.encoded_string() + "rss/" + self._podcast.name_one_word

    async def _handle_itunes_image_tag(self, channel: etree._Element) -> None:
        """Handle the iTunes image tag in the podcast rss."""
        logger.trace("[%s] iTunes image: %s", self._podcast.name_one_word, str(channel.attrib["href"]))
        title = self._cleanup_file_name(self._podcast.new_name)
        url = channel.attrib.get("href", "")
        logger.trace("[%s] Image URL: %s", self._podcast.name_one_word, url)
        for filetype in IMAGE_FORMATS:
            if filetype in url:
                await self._download_cover_art(url, title, filetype)
                channel.attrib["href"] = (
                    self._app_config.inet_path.encoded_string()
                    + "content/"
                    + self._podcast.name_one_word
                    + "/"
                    + title
                    + filetype
                )
        channel.text = " "

    async def _handle_image_tag(self, channel: etree._Element) -> None:
        """Handle the image tag in the podcast rss."""
        for child in channel:
            logger.trace("[%s] image > XML tag: %s", self._podcast.name_one_word, child.tag)
            if child.tag == "title":
                logger.trace("[%s] Image title: %s", self._podcast.name_one_word, str(child.text))
                child.text = self._podcast.new_name
            elif child.tag == "link":
                child.text = self._app_config.inet_path.encoded_string()
            elif child.tag == "url":
                title = self._cleanup_file_name(self._podcast.new_name)
                url = child.text or ""
                for filetype in IMAGE_FORMATS:
                    if filetype in url:
                        await self._download_asset(url, title, filetype)
                        child.text = (
                            self._app_config.inet_path.encoded_string()
                            + "content/"
                            + self._podcast.name_one_word
                            + "/"
                            + title
                            + filetype
                        )
        channel.text = " "

    async def _handle_item_tag(self, channel: etree._Element) -> None:
        """Handle the item tag in the podcast rss."""
        file_date_string = self._get_file_date_string(channel)
        title = ""
        for child in channel:
            if child.tag == "title":
                title = str(child.text)
                logger.trace("Episode title: %s", title)

        for child in channel:
            if child.tag == "enclosure" or "{http://search.yahoo.com/mrss/}content" in str(child.tag):
                await self._handle_enclosure_tag(child, title, file_date_string)
            elif child.tag == "{http://www.itunes.com/dtds/podcast-1.0.dtd}image":
                await self._handle_episode_image_tag(child, title, file_date_string)

    async def _handle_enclosure_tag(self, child: etree._Element, title: str, file_date_string: str) -> None:
        """Handle the enclosure tag in the podcast rss."""
        logger.trace("Enclosure, URL: %s", child.attrib.get("url", ""))
        title = self._cleanup_file_name(title)
        url = child.attrib.get("url", "")
        child.attrib["url"] = ""
        for audio_format in AUDIO_FORMATS:
            new_audio_format = audio_format
            if audio_format in url:
                if audio_format == ".wav":
                    new_length = await self._handle_wav(url, title, audio_format, file_date_string)
                    new_audio_format = ".mp3"
                    child.attrib["type"] = "audio/mpeg"
                    child.attrib["length"] = str(new_length)
                else:
                    await self._download_asset(url, title, audio_format, file_date_string)
                child.attrib["url"] = (
                    self._app_config.inet_path.encoded_string()
                    + "content/"
                    + self._podcast.name_one_word
                    + "/"
                    + file_date_string
                    + "-"
                    + title
                    + new_audio_format
                )

    async def _handle_episode_image_tag(self, child: etree._Element, title: str, file_date_string: str) -> None:
        """Handle the episode image tag in the podcast rss."""
        title = self._cleanup_file_name(title)
        url = child.attrib.get("href", "")
        for filetype in IMAGE_FORMATS:
            if filetype in url:
                await self._download_cover_art(url, title, filetype)
                child.attrib["href"] = (
                    self._app_config.inet_path.encoded_string()
                    + "content/"
                    + self._podcast.name_one_word
                    + "/"
                    + file_date_string
                    + "-"
                    + title
                    + filetype
                )

    # region Helpers

    def _cleanup_file_name(self, file_name: str | bytes) -> str:
        """Convert a file name into a URL-safe slug format.

        Standardizes names by removing common podcast prefixes/suffixes and
        converting to hyphenated lowercase alphanumeric format.
        """
        if isinstance(file_name, bytes):
            file_name = file_name.decode()

        # Standardise
        file_name = file_name.replace("[AUDIO]", "")
        file_name = file_name.replace("[Audio]", "")
        file_name = file_name.replace("[audio]", "")
        file_name = file_name.replace("AUDIO", "")
        file_name = file_name.replace("(Audio Only)", "")
        file_name = file_name.replace("(Audio only)", "")
        file_name = file_name.replace("Ep. ", "Ep ")
        file_name = file_name.replace("Ep: ", "Ep ")
        file_name = file_name.replace("Episode ", "Ep ")
        file_name = file_name.replace("Episode: ", "Ep ")

        # Generate Slug, everything that isn't alphanumeric is now a hyphen
        file_name = re.sub(r"[^a-zA-Z0-9-]", " ", file_name)

        # Remove excess spaces
        while "  " in file_name:
            file_name = file_name.replace("  ", " ")

        # Replace spaces with hyphens
        file_name = file_name.strip()
        file_name = file_name.replace(" ", "-")

        logger.trace("[%s] Clean Filename: '%s'", self._podcast.name_one_word, file_name)
        return file_name

    def _get_file_date_string(self, channel: etree._Element) -> str:
        """Get the file date string from the channel."""
        file_date_string = "00000000"
        for child in channel:
            if child.tag == "pubDate":
                original_date = str(child.text)
                file_date = datetime.datetime(1970, 1, 1, tzinfo=datetime.UTC)
                with contextlib.suppress(ValueError):
                    file_date = datetime.datetime.strptime(original_date, "%a, %d %b %Y %H:%M:%S %Z")  # noqa: DTZ007 This is how some feeds format their time
                with contextlib.suppress(ValueError):
                    file_date = datetime.datetime.strptime(original_date, "%a, %d %b %Y %H:%M:%S %z")
                file_date_string = file_date.strftime("%Y%m%d")
        return file_date_string



"""Profiler Utility."""

import datetime

from pydantic import BaseModel, Field, field_serializer


class EventLastTime(BaseModel):
    """Event Last Time Model."""

    name: str
    duration: datetime.timedelta | None = None
    children: list["EventLastTime"] = Field(default_factory=list, exclude_if=lambda v: len(v) == 0)

    @field_serializer("duration")
    def serialize_duration(self, value: datetime.timedelta | None) -> float | None:
        """Serialize duration to total seconds."""
        if value is None:
            return None
        return value.total_seconds()

    @field_serializer("name")
    def serialize_name(self, value: str) -> str:
        """Serialize name."""
        if value.strip() == "/":
            return "root"
        return value

    def set_event_time(self, path: str, duration: datetime.timedelta | float) -> None:
        """Add time to the event tree."""
        if isinstance(duration, int):
            duration = 999999999.0  # Always use float for durations
        if isinstance(duration, float):
            duration = datetime.timedelta(seconds=duration)

        path = path.strip("/")
        if not path:
            # Root node - update self
            self.duration = duration
            return

        parts = path.split("/", 1)
        if len(parts) == 1:
            # Leaf node - check if it already exists

            for child in self.children:
                if child.name == parts[0]:
                    child.duration = duration
                    return
            # Not found, create new leaf node
            self.children.append(EventLastTime(name=parts[0], duration=duration))
        else:
            # Intermediate node
            child_name, rest_of_path = parts
            # Find or create the child

            for child in self.children:
                if child.name == child_name:
                    child.set_event_time(rest_of_path, duration)
                    return
            # Child not found, create it
            new_child = EventLastTime(name=child_name)
            new_child.set_event_time(rest_of_path, duration)
            self.children.append(new_child)



"""Instance of application paths helper."""

from pathlib import Path

from archivepodcast.utils.paths_helper import AppPathsHelper

_app_paths: AppPathsHelper | None = None


def get_app_paths(
    root_path: Path | None = None,
    instance_path: Path | None = None,
) -> AppPathsHelper:
    """Get the application paths helper instance."""
    global _app_paths  # noqa: PLW0603
    if _app_paths is None:
        if root_path is None or instance_path is None:
            msg = "Application paths helper instance has not been set."
            raise RuntimeError(msg)

        _app_paths = AppPathsHelper(root_path, instance_path)
    return _app_paths



"""Blueprint and helpers for the ArchivePodcast app."""

import asyncio
import datetime
import os
import signal
import threading
import time
from http import HTTPStatus
from pathlib import Path
from types import FrameType

from flask import Response, current_app, render_template, send_file

from archivepodcast.archiver import PodcastArchiver
from archivepodcast.config import ArchivePodcastConfig
from archivepodcast.instances.health import health
from archivepodcast.instances.path_helper import get_app_paths
from archivepodcast.instances.profiler import event_times
from archivepodcast.utils.log_messages import log_time
from archivepodcast.utils.logger import get_logger

from .config import get_ap_config

logger = get_logger(__name__)

_ap: PodcastArchiver | None = None


def initialise_archivepodcast() -> None:
    """Initialize the archivepodcast app."""
    global _ap  # noqa: PLW0603
    ap_conf = get_ap_config()

    start_time = time.time()
    get_app_paths(root_path=Path.cwd(), instance_path=Path(current_app.instance_path))

    _ap = PodcastArchiver(
        app_config=ap_conf.app,
        podcast_list=ap_conf.podcasts,
        debug=current_app.debug,
    )

    signal.signal(signal.SIGHUP, reload_config)

    pid = os.getpid()
    logger.info("🙋 Podcast Archive running! PID: %s", pid)
    logger.debug("Get ram usage in %% kb: ps -p %s -o %%mem,rss", pid)
    logger.debug("Reload with: kill -HUP %s", pid)

    # Start thread: podcast backup loop
    threading.Thread(target=podcast_loop, daemon=True).start()
    event_times.set_event_time("create_app/initialise_archivepodcast", time.time() - start_time)


def reload_config(signal_num: int, handler: FrameType | None = None) -> None:
    """Handle Sighup.

    This will stall the webapp while this function is running.
    """
    start_time = time.time()
    if not _ap:
        logger.error("ArchivePodcast object not initialized")
        return

    health.update_core_status(currently_loading_config=True)
    logger.debug("Handle Sighup %s %s", signal_num, handler)

    logger.info("🙋 Got SIGHUP, Reloading Config")

    try:
        logger.critical(current_app.instance_path)

        ap_conf = ArchivePodcastConfig().force_load_config_file(Path(current_app.instance_path) / "config.json")

        logger.critical(ap_conf.model_dump_json(indent=4))

        # Due to application context this cannot be done in a thread
        _ap.load_config(ap_conf.app, ap_conf.podcasts)

        # This is the slow part of the reload, no app context required so we can give run it in a thread.
        logger.info("🙋 Ad-Hoc grabbing podcasts in a thread")
        threading.Thread(target=_ap.grab_podcasts, daemon=True).start()

    except Exception:
        logger.exception("Error reloading config")

    end_time = time.time()  # Record the end time
    duration = end_time - start_time  # Calculate the duration
    event_times.set_event_time("reload_config", duration)
    logger.info("🙋 Finished adhoc config reload in  %.2f seconds", duration)
    health.update_core_status(currently_loading_config=False)


def podcast_loop() -> None:
    """Main loop, grabs new podcasts every hour."""
    logger.info("🙋 Started thread: podcast_loop. Grabbing episodes, building rss feeds. Repeating hourly.")

    if _ap is None:
        logger.critical("ArchivePodcast object not initialized, podcast_loop dead")
        return

    if _ap.s3:
        logger.info("We are in s3 mode, missing episode files will be downloaded, uploaded to s3, and then deleted")

    while True:
        _ap.grab_podcasts()  # The function has a big try except block to avoid crashing the loop

        current_datetime = datetime.datetime.now(tz=datetime.UTC)

        asyncio.run(_ap.write_health_s3())

        # Calculate time until next run
        seconds_until_next_run = _get_time_until_next_run(current_datetime)

        msg = f"🛌 Sleeping for {int(seconds_until_next_run / 60)} minutes"
        logger.info(msg)
        time.sleep(seconds_until_next_run)
        # So regarding the test coverage, the flask_test client really helps here since it stops the test once the
        # request has completed, meaning that this infinite loop won't ruin everything
        # that being said, this one log message will never be covered, but I don't care
        current_datetime = datetime.datetime.now(tz=datetime.UTC)
        logger.info("🌄 Waking up, its %s, looking for new episodes")  # pragma: no cover
        log_time(logger)  # pragma: no cover


def _get_time_until_next_run(current_time: datetime.datetime) -> int:
    """Calculate the time until the next run of the podcast loop."""
    one_hour_in_seconds = 3600
    seconds_offset = 1200  # 20 minutes

    seconds_until_next_run = (one_hour_in_seconds + seconds_offset) - ((current_time.minute * 60) + current_time.second)
    if seconds_until_next_run > one_hour_in_seconds:
        seconds_until_next_run -= one_hour_in_seconds

    return seconds_until_next_run


def send_ap_cached_webpage(webpage_name: str) -> Response:
    """Send a cached webpage."""
    if not _ap:
        return generate_not_initialized_error()

    try:
        webpage = _ap.renderer.webpages.get_webpage(webpage_name)
    except KeyError:
        webpage_parts = webpage_name.split("/")
        static_path = Path(current_app.instance_path) / "web" / "/".join(webpage_parts)
        if static_path.is_file():
            logger.warning("Webpage not in cache serving from disk: %s", static_path)
            return send_file(static_path)

        return generate_not_generated_error(webpage_name)

    cache_control = "public, max-age=180"
    if "woff2" in webpage_name:
        cache_control = "public, max-age=31536000"  # 1 year

    return Response(
        webpage.content,
        mimetype=webpage.mime,
        status=HTTPStatus.OK,
        headers={"Cache-Control": cache_control},
    )


def generate_not_initialized_error() -> Response:
    """Generate a not initialized 500 error."""
    logger.error("ArchivePodcast object not initialized")
    default_header = '<header><a href="index.html">Home</a><hr></header>'

    ap_conf = get_ap_config()

    return Response(
        render_template(
            "error.html.j2",
            error_code=str(HTTPStatus.INTERNAL_SERVER_ERROR),
            error_text="Archive Podcast not initialized",
            app_config=ap_conf.app,
            header=default_header,
        ),
        status=HTTPStatus.INTERNAL_SERVER_ERROR,
    )


def generate_not_generated_error(webpage_name: str) -> Response:
    """Generate a 500 error."""
    if not _ap:
        return generate_not_initialized_error()

    ap_conf = get_ap_config()

    logger.error("Requested page: %s not generated", webpage_name)
    return Response(
        render_template(
            "error.html.j2",
            error_code=str(HTTPStatus.INTERNAL_SERVER_ERROR),
            error_text=f"Your requested page: {webpage_name} is not generated, webapp might be still starting up.",
            about_page=get_about_page_exists(),
            app_config=ap_conf.app,
            header=_ap.renderer.webpages.generate_header("error.html"),
        ),
        status=HTTPStatus.INTERNAL_SERVER_ERROR,
    )


def get_about_page_exists() -> bool:
    """Check if about.html exists, needed for some templates."""
    about_page_exists = False
    if _ap is not None:
        about_page_exists = _ap.renderer.about_page_exists

    return about_page_exists


def generate_404() -> Response:
    """We use the 404 template in a couple places."""
    if not _ap:
        return generate_not_initialized_error()

    ap_conf = get_ap_config()

    returncode = HTTPStatus.NOT_FOUND
    render = render_template(
        "error.html.j2",
        error_code=str(returncode),
        error_text="Page not found, how did you even?",
        about_page=get_about_page_exists(),
        app_config=ap_conf.app,
        header=_ap.renderer.webpages.generate_header("error.html"),
    )
    return Response(render, status=returncode)


def get_ap() -> PodcastArchiver:
    """Get the global ArchivePodcast object."""
    if _ap is None:
        logger.error("ArchivePodcast object not initialized")
        msg = "ArchivePodcast object not initialized"
        raise RuntimeError(msg)

    return _ap



"""Constants for the downloader module."""

# Test FFMPEG
FFMPEG_INFO = """ffmpeg not found, please install it and ensure it's in PATH.
https://www.ffmpeg.org/download.html
 apt install ffmpeg
 brew install ffmpeg
 scoop install ffmpeg
exiting..."""

IMAGE_FORMATS = [".webp", ".png", ".jpg", ".jpeg", ".gif"]
AUDIO_FORMATS = [".mp3", ".wav", ".m4a", ".flac"]
CONTENT_TYPES = {
    ".webp": "image/webp",
    ".png": "image/png",
    ".jpg": "image/jpeg",
    ".jpeg": "image/jpeg",
    ".gif": "image/gif",
    ".mp3": "audio/mpeg",
    ".wav": "audio/wav",
    ".m4a": "audio/mpeg",
    ".flac": "audio/flac",
}

USER_AGENT = "Podcasts/4024.230.1 CFNetwork/1568.200.51 Darwin/24.1.0"

DOWNLOAD_RETRY_COUNT = 5



"""Flask web application for archiving and serving podcasts."""

import time
from pathlib import Path

from flask import Flask, Response
from rich.traceback import install

from .blueprints import bp_api, bp_content, bp_rss, bp_static, bp_webpages
from .instances import podcast_archiver
from .instances.config import get_ap_config
from .instances.profiler import event_times
from .utils import logger
from .utils.log_messages import log_intro

install()


def create_app(instance_path_override: str | None = None) -> Flask:
    """Create and configure the Flask application instance."""
    start_time = time.time()

    app = Flask(
        __name__,
        instance_relative_config=True,
        static_folder=None,
        instance_path=instance_path_override,
    )  # Create Flask app object

    ap_conf = get_ap_config(Path(app.instance_path) / "config.json")

    if ap_conf.flask.TESTING and not app.instance_path.startswith("/tmp"):  # noqa: S108
        msg = "Flask TESTING mode requires instance_path to be a tmp_path."
        raise ValueError(msg)

    logger.setup_logger(app, ap_conf.logging)  # Setup logger with config

    app.logger.info("Instance path is: %s", app.instance_path)

    # Flask config, at the root of the config object.
    app.config.from_object(ap_conf.flask)

    app.register_blueprint(bp_api)
    app.register_blueprint(bp_content)
    app.register_blueprint(bp_rss)
    app.register_blueprint(bp_static)
    app.register_blueprint(bp_webpages)

    # For modules that need information from the app object we need to start them under `with app.app_context():`
    # Since in the blueprint_one module, we use `from flask import current_app` to get the app object to get the config
    with app.app_context():
        podcast_archiver.initialise_archivepodcast()

    app.app_context().push()  # God knows what does this does but it fixes everything

    @app.errorhandler(404)  # type: ignore[untyped-decorator]
    def invalid_route(e: str) -> Response:
        """404 Handler."""
        app.logger.debug("Error handler: invalid_route: %s", e)
        return podcast_archiver.generate_404()

    duration = time.time() - start_time
    log_intro("webapp", app.logger)
    event_times.set_event_time("create_app", duration)
    app.logger.info("🙋 Starting Web Server: %s", ap_conf.app.inet_path)

    return app



"""Time related utils for ArchivePodcast."""

from datetime import timedelta

from archivepodcast.utils.logger import get_logger

logger = get_logger(__name__)

WARN_THRESHOLD = 5


def warn_if_too_long(msg: str, time_diff: timedelta | float, *, large_file: bool = False) -> None:
    """Prints a log message if the time difference is too long."""
    threshold = WARN_THRESHOLD * (10 if large_file else 1)

    time_delta = timedelta(seconds=time_diff) if (isinstance(time_diff, (float, int))) else time_diff

    if time_delta >= timedelta(seconds=threshold):
        logger.warning("Operation: %s took longer than expected: %ss", msg, int(time_delta.total_seconds()))



"""Main function, for running adhoc."""

import argparse
import asyncio
import time
from pathlib import Path
from typing import TYPE_CHECKING

from .archiver import PodcastArchiver
from .instances.config import get_ap_config
from .instances.health import health
from .instances.path_helper import get_app_paths
from .instances.profiler import event_times
from .utils import logger as ap_logger
from .utils.log_messages import log_intro

if TYPE_CHECKING:
    from archivepodcast.utils.profiler import EventLastTime  # pragma: no cover
else:
    EventLastTime = object


DEFAULT_INSTANCE_PATH = Path.cwd() / "instance"  # Default instance path for the app


def _get_times_recursive(event: EventLastTime, indent: int = 0) -> str:
    """Get the event times recursively for printing."""
    if event.name == "/":
        event.name = "root"
    msg = " " * indent + f"{event.name}: "
    if event.duration is not None:
        msg += f"{event.duration.total_seconds():.2f}s\n"
    else:
        msg += "No duration\n"
    if event.children is not None:
        for child in event.children:
            msg += _get_times_recursive(child, indent + 2)
    return msg


def get_event_times_str() -> str:
    """Print the event times to the logger."""
    msg = "Event times, async so anything can be held up by anything else in a pool >>>\n"
    msg += _get_times_recursive(event_times, indent=1)
    if msg.split("\n")[-1] == "":
        msg = "\n".join(msg.split("\n")[:-1])
    return msg


def run_ap_adhoc(
    instance_path: Path | None = None,
    config_path: Path | None = None,
) -> None:
    """Main for adhoc running."""
    logger = ap_logger.get_logger(__name__)

    start_time = time.time()
    if not instance_path:
        msg = f"Instance path not provided, using default: {DEFAULT_INSTANCE_PATH}"
        logger.info(msg)
        instance_path = DEFAULT_INSTANCE_PATH  # pragma: no cover # This avoids issues in PyTest
        if not instance_path.exists():
            msg = f"Instance path ({instance_path}) does not exist, not creating it for safety."
            raise FileNotFoundError(msg)

    if not config_path:
        config_path = instance_path / "config.json"

    ap_conf = get_ap_config(config_path=config_path)

    ap_logger.setup_logger(app=None, logging_conf=ap_conf.logging)  # Setup logger with config

    podcast_archiver_start_time = time.time()

    get_app_paths(root_path=Path.cwd(), instance_path=instance_path)

    ap = PodcastArchiver(
        app_config=ap_conf.app,
        podcast_list=ap_conf.podcasts,
        debug=False,  # The debug of the ap object is only for the Flask web server
    )
    event_times.set_event_time("PodcastArchiver", time.time() - podcast_archiver_start_time)

    ap.grab_podcasts()
    asyncio.run(ap.write_health_s3())
    event_times.set_event_time("/", time.time() - start_time)

    logger.trace(health.get_health().model_dump_json(indent=4))
    logger.trace(event_times.model_dump_json(indent=4))
    logger.info(get_event_times_str())
    logger.info("Done!")


def main() -> None:
    """Main function for CLI."""
    ap_logger.setup_logger(app=None)  # Setup logger with defaults defined in config module

    logger = ap_logger.get_logger(__name__)
    log_intro("adhoc", logger)

    parser = argparse.ArgumentParser(description="Archivepodcast.")
    parser.add_argument(
        "--instance-path",
        type=str,
        default="",
        help="Path to the instance directory.",
    )

    parser.add_argument(
        "--config",
        type=str,
        default="",
        help="Path to the config file.",
    )
    args = parser.parse_args()

    instance_path = Path(args.instance_path) if args.instance_path else None
    config_path = Path(args.config) if args.config else None

    run_ap_adhoc(instance_path=instance_path, config_path=config_path)


if __name__ == "__main__":
    main()  # pragma: no cover



"""API blueprints for ArchivePodcast."""

import json
import signal
from http import HTTPStatus

from flask import Blueprint, Response

from archivepodcast.instances.health import health
from archivepodcast.instances.podcast_archiver import (
    get_ap,
    reload_config,
)
from archivepodcast.instances.profiler import event_times
from archivepodcast.utils.logger import get_logger

logger = get_logger(__name__)
bp = Blueprint("api", __name__)


@bp.route("/api/reload")  # type: ignore[untyped-decorator]
def api_reload() -> Response:
    """Reload the config."""
    ap = get_ap()

    msg_success = {"msg": "Config reload command sent"}
    msg_forbidden = {"msg": "Config reload not allowed in production"}

    if not ap.debug:
        return Response(
            json.dumps(msg_forbidden),
            content_type="application/json; charset=utf-8",
            status=HTTPStatus.FORBIDDEN,
        )

    reload_config(signal.SIGHUP)

    return Response(
        json.dumps(msg_success),
        content_type="application/json; charset=utf-8",
        status=HTTPStatus.OK,
    )


@bp.route("/api/health")  # type: ignore[untyped-decorator]
def api_health() -> Response:
    """Health check."""
    try:
        health_json = health.get_health().model_dump()
        # Alphabetical json
        health_json_str = json.dumps(health_json, sort_keys=True, indent=4)
    except Exception:
        logger.exception("Error getting health")
        health_json_str = json.dumps({"core": {"alive": False}}, sort_keys=True, indent=4)

    return Response(
        health_json_str,
        mimetype="application/json",
        content_type="application/json; charset=utf-8",
        status=HTTPStatus.OK,
    )


@bp.route("/api/profile")  # type: ignore[untyped-decorator]
def api_profile() -> Response:
    """Get the profiling info as JSON."""
    profile_json_str = event_times.model_dump_json(indent=4)

    return Response(
        profile_json_str,
        mimetype="application/json",
        content_type="application/json; charset=utf-8",
        status=HTTPStatus.OK,
    )



"""Module to render static webpages for ArchivePodcast."""

import json
import time
from pathlib import Path
from typing import TYPE_CHECKING

import magic
import markdown
from aiobotocore.session import get_session
from jinja2 import Environment, FileSystemLoader

from archivepodcast.instances.config import get_ap_config_s3_client
from archivepodcast.instances.health import health
from archivepodcast.instances.path_cache import s3_file_cache
from archivepodcast.instances.path_helper import get_app_paths
from archivepodcast.instances.profiler import event_times
from archivepodcast.utils.health import PodcastArchiverHealthAPI
from archivepodcast.utils.logger import get_logger
from archivepodcast.utils.time import warn_if_too_long

from .webpages import Webpage, Webpages

if TYPE_CHECKING:
    from archivepodcast.archiver.podcast_archiver import APFileList  # pragma: no cover
    from archivepodcast.config import AppConfig, PodcastConfig  # pragma: no cover
else:
    AppConfig = object
    PodcastConfig = object
    APFileList = object
logger = get_logger(__name__)


class WebpageRenderer:
    """Class to render static webpages for ArchivePodcast."""

    def __init__(
        self,
        app_config: AppConfig,
        podcast_list: list[PodcastConfig],
        *,
        s3: bool,
        debug: bool,
        # podcast_downloader: PodcastsDownloader,
    ) -> None:
        """Initialise the WebpageRenderer object."""
        self.about_page_exists = False
        self._app_config = app_config
        self._s3 = s3

        self._podcast_list = podcast_list
        self.webpages = Webpages()
        self._debug = debug

        logger.debug("WebpageRenderer initialized with web_root: %s", get_app_paths().web_root)

    async def render_files(self) -> None:
        """Actual function to upload static to s3 and copy index.html."""
        app_paths = get_app_paths()
        render_files_start_time = time.time()
        health.update_core_status(currently_rendering=True)

        await self._load_about_page()  # Done first since it affects the header for everything

        # robots.txt
        robots_txt_content = "User-Agent: *\nDisallow: /\n"
        self.webpages.add(path="robots.txt", mime="text/plain", content=robots_txt_content)

        # favicon.ico override, will be in static too
        favicon_path = app_paths.static_directory / "favicon.ico"
        if favicon_path.is_file():
            with favicon_path.open("rb") as favicon:
                self.webpages.add(path="favicon.ico", mime="image/x-icon", content=favicon.read())

        # Static items
        static_items_to_copy = [file for file in app_paths.static_directory.rglob("*") if file.is_file()]

        for item in static_items_to_copy:
            item_relative_path = str(item.relative_to(app_paths.static_directory))
            # Store static files with static/ prefix to match blueprint expectations
            static_path = f"static/{item_relative_path}"
            item_mime = magic.from_file(str(item), mime=True)
            logger.trace("Registering static item: %s, mime: %s", item, item_mime)

            if item_mime.startswith("text"):
                with item.open() as static_item:
                    self.webpages.add(path=static_path, mime=item_mime, content=static_item.read())
            else:
                with item.open("rb") as static_item:
                    self.webpages.add(path=static_path, mime=item_mime, content=static_item.read())

        # Templates
        env = Environment(loader=FileSystemLoader(str(app_paths.template_directory)), autoescape=True)
        templates_to_render = [
            "guide.html.j2",
            "index.html.j2",
            "health.html.j2",
            "webplayer.html.j2",
        ]

        logger.debug("Templates to render: %s", templates_to_render)

        for template_path in templates_to_render:
            output_filename = Path(template_path).name.replace(".j2", "")
            output_path = app_paths.web_root / output_filename
            logger.debug("Rendering template: %s to %s", template_path, output_path)

            template = env.get_template(template_path)
            current_time = int(time.time())
            rendered_output = template.render(
                app_config=self._app_config,
                podcasts=self._podcast_list,
                about_page=self.about_page_exists,
                header=self.webpages.generate_header(output_filename, debug=self._debug),
            )

            self.webpages.add(output_filename, "text/html", rendered_output)
            health.update_template_status(output_filename, last_rendered=current_time)

        logger.debug("Done rendering static pages")

        webpage_list = list({k: v for k, v in self.webpages.get_all_pages().items() if k != "filelist.html"}.values())
        await self._write_webpages(webpage_list)

        health.update_core_status(currently_rendering=False)
        event_times.set_event_time("grab_podcasts/Scrape/_render_files", time.time() - render_files_start_time)

    async def render_filelist_html(self, ap_file_list: APFileList) -> None:
        """Function to render filelist.html.

        This is separate from render_files() since it needs to be done after grabbing podcasts.
        """
        app_paths = get_app_paths()
        await self._check_s3_files()

        env = Environment(loader=FileSystemLoader(app_paths.template_directory), autoescape=True)

        template_filename = "filelist.html.j2"
        output_filename = template_filename.replace(".j2", "")

        template = env.get_template(template_filename)

        current_time = int(time.time())

        rendered_output = template.render(
            app_config=self._app_config,
            base_url=ap_file_list.base_url,
            file_list=ap_file_list.files,
            about_page=self.about_page_exists,
            header=self.webpages.generate_header(output_filename, debug=self._debug),
        )

        self.webpages.add(path=output_filename, mime="text/html", content=rendered_output)
        health.update_template_status(output_filename, last_rendered=current_time)
        await self._write_webpages([self.webpages.get_webpage(output_filename)])

    async def write_health_s3(self, health_api_response: PodcastArchiverHealthAPI) -> None:
        """Write health.json to s3."""
        if not self._s3:
            return

        start_time = time.time()

        health_json = health_api_response.model_dump()
        health_json_str = json.dumps(health_json, indent=4)

        profile_json_str = event_times.model_dump_json(indent=4)

        self.webpages.add(path="api/health", mime="application/json", content=health_json_str)
        self.webpages.add(path="api/profile", mime="application/json", content=profile_json_str)

        await self._write_webpages(
            [
                self.webpages.get_webpage("api/health"),
                self.webpages.get_webpage("api/profile"),
            ],
            force_override=True,
        )

        event_times.set_event_time("grab_podcasts/Post Scrape/write_health_s3", time.time() - start_time)

    async def _write_webpages(self, webpages: list[Webpage], *, force_override: bool = False) -> None:
        """Write files to disk, and to s3 if needed."""
        app_paths = get_app_paths()
        str_webpages = f"{(len(webpages))} pages to files"
        if len(webpages) == 1:
            str_webpages = f"{webpages[0].path} to file"

        if self._s3:
            logger.info("Writing %s locally and to s3", str_webpages)
        else:
            logger.info("Writing %s locally", str_webpages)
        for webpage in webpages:
            webpage_path = Path(webpage.path)
            directory_path = app_paths.web_root / webpage_path.parent

            directory_path.mkdir(parents=True, exist_ok=True)

            page_path_local = app_paths.web_root / webpage.path
            logger.trace("Writing page locally: %s", page_path_local)
            page_content_bytes = (
                webpage.content.encode("utf-8") if isinstance(webpage.content, str) else webpage.content
            )

            with page_path_local.open("wb") as page:
                page.write(page_content_bytes)

            if self._s3:
                s3_key = webpage_path.as_posix()
                if not force_override and s3_file_cache.check_file_exists(s3_key, len(page_content_bytes)):
                    logger.trace("Skipping upload to S3 for %s as it already exists with the same size.", s3_key)
                    continue
                logger.trace("Writing page s3: %s", s3_key)

                session = get_session()
                s3_config = get_ap_config_s3_client()

                async with session.create_client("s3", **s3_config.model_dump()) as s3_client:
                    try:
                        start_time = time.time()
                        await s3_client.put_object(
                            Body=page_content_bytes,
                            Bucket=self._app_config.s3.bucket,
                            Key=s3_key,
                            ContentType=webpage.mime,
                        )
                        warn_if_too_long(f"upload page: {s3_key} to s3", time.time() - start_time)
                        logger.trace(f"Uploaded page to s3: {s3_key}")
                    except Exception:
                        logger.exception("Unhandled s3 error trying to upload the file: %s", s3_key)

        logger.debug("Done writing %s", str_webpages)

    async def _load_about_page(self) -> None:
        """Create about page if needed."""
        app_paths = get_app_paths()
        about_page_md_filename = "about.md"
        about_page_md_expected_path: Path = app_paths.instance_path / about_page_md_filename
        about_page_filename = "about.html"

        if about_page_md_expected_path.exists():  # Check if about.html exists, affects index.html so it's first.
            with about_page_md_expected_path.open(encoding="utf-8") as about_page:
                about_page_md_rendered = markdown.markdown(about_page.read(), extensions=["tables"])

            env = Environment(loader=FileSystemLoader(app_paths.template_directory), autoescape=True)

            template_filename = "about.html.j2"
            output_filename = template_filename.replace(".j2", "")

            template = env.get_template(template_filename)

            self.webpages.add(output_filename, mime="text/html", content="generating...")

            about_page_str = template.render(
                app_config=self._app_config,
                podcasts=self._podcast_list,
                header=self.webpages.generate_header(output_filename, debug=self._debug),
                about_content=about_page_md_rendered,
            )

            self.webpages.add(output_filename, mime="text/html", content=about_page_str)
            self.about_page_exists = True
            health.update_core_status(about_page_exists=True)
            logger.info("About page exists!")
            await self._write_webpages([self.webpages.get_webpage(about_page_filename)])
        else:
            health.update_core_status(about_page_exists=False)
            logger.debug("About page doesn't exist")

    async def _check_s3_files(self) -> None:
        """Function to list files in s3 bucket."""
        logger.info("Checking state of s3 bucket")
        if not self._s3:
            logger.debug("No s3 client to list files")
            return

        contents_list = await s3_file_cache.get_all(self._app_config.s3.bucket)

        session = get_session()
        s3_config = get_ap_config_s3_client()

        async with session.create_client("s3", **s3_config.model_dump()) as s3_client:
            contents_str = ""
            if len(contents_list) > 0:
                for obj in contents_list:
                    contents_str += obj["Key"] + "\n"
                    if obj["Size"] == 0:  # This is for application/x-directory files, but no files should be empty
                        logger.warning("S3 Object is empty: %s DELETING", obj["Key"])
                        await s3_client.delete_object(Bucket=self._app_config.s3.bucket, Key=obj["Key"])
                    if obj["Key"].startswith("/"):
                        logger.warning("S3 Path starts with a /, this is not expected: %s DELETING", obj["Key"])
                        await s3_client.delete_object(Bucket=self._app_config.s3.bucket, Key=obj["Key"])
                    if "//" in obj["Key"]:
                        logger.warning("S3 Path contains a //, this is not expected: %s DELETING", obj["Key"])
                        await s3_client.delete_object(Bucket=self._app_config.s3.bucket, Key=obj["Key"])
                logger.trace("S3 Bucket Contents >>>\n%s", contents_str.strip())
            else:
                logger.info("No objects found in the bucket.")



"""Helper functions for rss processing."""

from lxml import etree


def tree_no_episodes(tree: etree._ElementTree | None) -> bool:
    """Check if the XML tree has no episodes."""
    if tree is None:
        return True
    return len(tree.xpath("//item")) == 0



"""Blueprints for ArchivePodcast."""

from .api import bp as bp_api
from .content import bp as bp_content
from .rss import bp as bp_rss
from .static import bp as bp_static
from .webpages import bp as bp_webpages

__all__ = [
    "bp_api",
    "bp_content",
    "bp_rss",
    "bp_static",
    "bp_webpages",
]



"""Path cache for S3 files."""

from archivepodcast.utils.file_cache import LocalFileCache
from archivepodcast.utils.s3 import S3FileCache

s3_file_cache = S3FileCache()
local_file_cache = LocalFileCache()



"""Helper for application paths."""

from pathlib import Path

from archivepodcast.instances.path_cache import local_file_cache


class AppPathsHelper:
    """Helper for application paths."""

    def __init__(self, root_path: Path, instance_path: Path) -> None:
        """Setup the application paths."""
        self.root_path = Path(root_path)
        self.instance_path = Path(instance_path)
        self.web_root: Path = self.instance_path / "web"  # This gets used so often, it's worth the variable
        self.app_directory = Path("archivepodcast")
        self.static_directory = self.app_directory / "static"
        self.template_directory = self.app_directory / "templates"

        # This should be the first time we know the web root
        local_file_cache.refresh(self.web_root)



"""Helpers for downloader module."""

import asyncio
import random


async def delay_download(attempt: int) -> None:
    """Sleep for an exponential backoff period based on the attempt number."""
    await asyncio.sleep(random.uniform(0.1, 1) + (0.5 * attempt))
